@@ -17,8 +17,10 @@
 
 from __future__ import annotations
 
+import json
 import logging
 from typing import cast
+from urllib.parse import quote
 
 from fastapi import Request  # noqa: TC002
 from fastapi.responses import HTMLResponse, RedirectResponse
@@ -72,7 +74,10 @@ def login_callback(request: Request):
         code=code,
         redirect_uri=str(redirect_uri),
     )
-    userinfo = client.userinfo(tokens["access_token"])
+    userinfo_raw: dict | bytes = client.userinfo(tokens["access_token"])
+    # Decode bytes to dict if necessary
+    userinfo: dict = json.loads(userinfo_raw) if isinstance(userinfo_raw, bytes) else userinfo_raw
+
     user = KeycloakAuthManagerUser(
         user_id=userinfo["sub"],
         name=userinfo["preferred_username"],
@@ -108,7 +113,11 @@ def logout(request: Request):
     post_logout_redirect_uri = request.url_for("logout_callback")
 
     if id_token:
-        logout_url = f"{end_session_endpoint}?post_logout_redirect_uri={post_logout_redirect_uri}&id_token_hint={id_token}"
+        encoded_id_token = quote(id_token, safe="")
+        logout_url = (
+            f"{end_session_endpoint}?post_logout_redirect_uri={post_logout_redirect_uri}"
+            f"&id_token_hint={encoded_id_token}"
+        )
     else:
         logout_url = str(post_logout_redirect_uri)
 
